# terminal styles for better logging
import sys
from enum import Enum
from colorama import Fore, Back, Style, init as init_term, AnsiToWin32
from termcolor import colored
# colorful terminal initializacion
init_term(wrap=True, autoreset=True)
stream = AnsiToWin32(sys.stderr).stream


class LogLevel(Enum):
    DEBUG = 'DEBUG'  # Blue
    ERROR = 'ERROR'  # Red
    INFO = 'INFO'   # White
    WARNING = 'WARNING'  # Yellow
    SUCCESS = 'SUCCESS'  # ! Green, Non-standard with Eliot (might not care)


class LogStyle(Enum):
    PREFIX_FIRST_LINE = 'PREFIX_FIRST_LINE'
    PREFIX_ALL_LINES = 'PREFIX_ALL_LINES'

# print wrappers


def log(level: LogLevel, message: str):
    """Logs formatted messages to stdout

    Parameters
    ----------
    level : LogLevel
        Message Level and color related to it (DEBUG, ERROR, etc, see LogLevel type)
    message : Any
        Message to log
    """
    if(level == LogLevel.INFO):
        log_info(message)
    if(level == LogLevel.SUCCESS):
        log_success(message)
    if(level == LogLevel.WARNING):
        log_warning(message)
    if(level == LogLevel.ERROR):
        log_error(message)
    if(level == LogLevel.DEBUG):
        log_debug(message)


def log_info(message):
    print(colored(message, 'white'))


def log_success(message):
    print(colored(message, 'green'))


def log_warning(message):
    print(colored(message, 'yellow'))


def log_error(message):
    print(colored(message, 'red'))


def log_debug(message):
    print(colored(message, 'blue'))


'''
# Multiprocessing Pipes reference (For inter-process communication):

# import multiprocessing as mp
# from multiprocessing import Pipe
# import time
# context = mp.get_context('fork')

# def fn_slow():
#   count = 5
#   while count:
#     count -= 1
#     time.sleep(1.0)

# pslow = context.Process(target=fn_slow)
# pslow.start()


kill = False
def fn(id, conn):
  started = time.time()
  while not kill:
    time.sleep(5.0)
    conn.send({
      'id': id,
      'level': 'DEBUG',
      'msg': f'Tick: {time.time() - started}'
    })

ps = []

CONN, conn = Pipe()
ps.append({
  'p': context.Process(target=fn, args=(len(ps)+1, conn,)),
  'c': CONN
})

ps[-1]['p'].start()
ps[-1]['p'].pid

try:
  while 1:
    _conn = ps[-1]['c']
    if _conn.poll():
      print(_conn.recv())
    time.sleep(0.1)
except KeyboardInterrupt:
  kill = True

ps[-1]['p'].join()

'''



import pydash as __
from pprint import pformat
import time
import json
from typing import Callable, Dict
from json.decoder import JSONDecodeError
import multiprocessing as mp
from multiprocessing import SimpleQueue, Pipe
from datetime import datetime, timedelta
from enum import Enum
from pygqlc import GraphQLClient
from singleton_decorator import singleton
from .Notifications import NotificationBehaviour
from .Logging import log, LogLevel, LogStyle
from . import uploaders
from . import queries
from . import mutations
from . import subscriptions
from .dateHelpers import getUtcDate
from .croniterHelpers import get_croniter
from .redis import WorkerRedis


class QueueType(Enum):
    """Enumeration for all the queue types
    """
    EVENT = 'EVENT'
    FREQUENCY = 'FREQUENCY'
    ON_DEMAND = 'ON_DEMAND'
    SCHEDULE = 'SCHEDULE'
    CONTINUOUS = 'CONTINUOUS'


class JobStatus(Enum):
    """Enumeration for all the job status.
    """
    ABORTED = 'ABORTED'
    ERROR = 'ERROR'
    FINISHED = 'FINISHED'
    PAUSED = 'PAUSED'
    RUNNING = 'RUNNING'
    WAITING = 'WAITING'


class PollingMode(Enum):
    """Set all the polling modes
    """
    QUERY = 'QUERY'
    SUBSCRIPTION = 'SUBSCRIPTION'


class QueryOrderBy(Enum):
    """Set Query Filert OrderBy
    """
    ID = 'ID'
    INSERTED_AT = 'INSERTED_AT'


class JobConfigMode(Enum):
    """Enumeration for all the job's configurations
    """
    KEEP = 'KEEP'  # do not touch anything at startup
    SYNC = 'SYNC'  # sync Queues, job configuration, remote logging configuration, etc


DEFAULT_EVENT_CONFIG = {
    "timeout": 5.0,  # seconds to wait, None to run ASAP
    "min_batch": 10,
    "max_batch": 100
}


def safe_pop(data, index=0, default=None):
    """This function makes a safe pop of a list.

    Args:
        data (list): List that want to make a pop
        index (int, optional): Element that want to be poped. Defaults to 0.
        default (optional): Default return if the element doesn't exists. Defaults to None.

    Returns:
        element list: Value of the element poped.
    """
    if len(data) > 0:
        return data.pop(index)
    else:
        return default


@singleton
class ValiotWorker():
    """The ValiotWorker class follows the singleton design pattern. It
      manages all listener logic for any action of a Graphql schema for many queues.

      Attributes:
          queues (dict): Dictonary with all queues. Defaults to
            empty dict.
          eventQueues (dict): Dictonary with all event queues. Defaults to
            empty dict.
          currentQueues (list): List with all active queues.
            Defaults to empty list.
          previousQueues (list): List with all previous queues.
            Defaults to empty list.
          runningJobs (dict): Dictonary with all running jobs. Defaults to
            empty dict.
          lockedJobs (list): List with all locked jobs.
            Defaults to empty list.
          gql (GraphQLClient): GraphQLClient instance.
          worker (string): Worker name.  Defaults to None.
          worker_code (string): Worker id.  Defaults to None.
          pollingMode (PollingMode): Set how will be the polling mode. Defaults to
            Query mode.
          jobConfigMode (JobConfigMode): Set the job configuration. Defaults to KEEP.
          context (Process): Context of a run job. Defaults to None.
          unsubJobCreated (GraphQLClient.sub): GraphQLClient subscriptions. Defaults
            to None.
          useRedis (boolean): Determine to use Redis. Defaults to False.
          redis (WorkerRedis): Instance of a WorkerRedis.
          continuousJobs (list): List of a running job. Defaults to list.
          logginStyle (LogStyle.PREFIX_FIRST_LINE) = LogStyle.PREFIX_FIRST_LINE
          stopLogging (boolean): Set to stop logs.
          loggingProcess (Process): Gets the logs of a process. Defaults to None.
          log (Log): Log message. Defaults to None.
          logQueues (dict): A dictionry with all queues logs.

      Examples:
          >>> <With> clause:
            '''
            import os
            from ValiotWorker import ValiotWorker, PollingMode
            from ValiotWorker.Logging import LogStyle
            from .config import setup_gql
            gql = setup_gql()
            # ! The worker must initialize before Job imports
            worker = ValiotWorker()
            worker.setClient(gql)
            worker.setWorker(os.environ.get('WORKER'))
            worker.setPollingMode(PollingMode.SUBSCRIPTION)
            worker.setLoggingStyle(LogStyle.PREFIX_FIRST_LINE) # default, just for reference
            # ! Import jobs for the worker to recognize them
            from .services.jobs.testComplexOnDemand.__main__ import testComplexOnDemand
            from .services.jobs.testOnDemand.__main__ import testOnDemand
            from .services.jobs.testDirectedNotification.__main__ import testDirectedNotification
            from .services.jobs.testGroupNotification.__main__ import testGroupNotification
            from .services.jobs.testOnEvent.__main__ import testOnEvent
            def main ():
              print('Test suite for Valiot jobs!')
              worker.run(interval=1.0)

            if __name__ == "__main__":
              main()
            '''
    """

    def __init__(self):
        self.queues = {}
        self.eventQueues = {}
        self.currentQueues = []
        self.previousQueues = []
        self.runningJobs = {}
        self.lockedJobs = []
        self.gql = GraphQLClient()
        self.worker = None
        self.worker_code = None
        self.pollingMode = PollingMode.QUERY
        self.queryOrderBy = QueryOrderBy.ID
        self.jobConfigMode = JobConfigMode.KEEP
        self.context = None
        self.unsubJobCreated = None
        self.useRedis = False
        self.redis = WorkerRedis()
        self.continuousJobs = []
        # logging attributes
        self.loggingStyle = LogStyle.PREFIX_FIRST_LINE
        self.stopLogging = False
        self.loggingProcess = None
        self.log = None
        self.logQueues = {
            'main': SimpleQueue()  # queue for the main thread (Orchestator)
        }

    def setClient(self, client):
        """Set a GraphQLClient instance.

        Args:
            client (GraphQLClient): GraphQLClient instance.
        """
        self.gql = client

    def setWorker(self, worker):
        """Sets a worker code.

        Args:
            worker (string): Worker code.
        """
        self.worker_code = worker

    def setPollingMode(self, mode):
        """Sets Polling Mode

        Args:
            mode (PollingMode): PollingMode option.
        """
        self.pollingMode = mode

    def setQueryOrderBy(self, orderBy):
        """Sets QueryOrderBy option

        Args:
            orderBy (QueryOrderBy): QueryOrderBy option.
        """
        self.queryOrderBy = orderBy

    def setJobConfigMode(self, mode):
        """Sets Jobs configuration.

        Args:
            mode (JobConfigMode): JobConfigMode option.
        """
        self.jobConfigMode = mode

    def setLoggingStyle(self, style):
        """Sets logging style.

        Args:
            style (LoggingStyle): LoggingStyle option.
        """
        self.loggingStyle = style

    def setUseRedis(self):
        """Set that worker uses Redis.
        """
        self.useRedis = True

    def job(
        self,
        name,
        alias,
        description='',
        schedule='',
        enabled=None,  # ! PER-WORKER BASIS
        queueType=QueueType.ON_DEMAND,
        query='',
        lockRequired=False,
        notificationBehaviour=NotificationBehaviour.DEFAULT,
        notificationFrequency='',
        notificationSchedule='',
        eventConfig=DEFAULT_EVENT_CONFIG
    ):
        """This fuction is the decorator for a job.

        Args:
            name (string): Job's name.
            alias (string): Job's alias.
            description (str, optional): Job's description. Defaults to ''.
            schedule (str, optional): Job's schedule in cron like format.
              Defaults to ''.
            enabled (boolean, optional): Set if the job is avaible. Defaults to None.
            query (str, optional): Query that will follow the job. Defaults to ''.
            lockRequired (bool, optional): Option that the job can lock.
              Defaults to False.
            notificationBehaviour (NotificationBehaviour, optional): Notification
              behaviour. Defaults to NotificationBehaviour.DEFAULT.
            notificationFrequency (str, optional): Notification frequency Defaults to ''.
            notificationSchedule (str, optional): Notification Schedule in cron like
              format. Defaults to ''.
            eventConfig (optional): Event configuration. Defaults to
              DEFAULT_EVENT_CONFIG.
        """
        def wrap(f):
            parameters = {
                'name': name,
                'alias': alias,
                'description': description,
                'schedule': schedule,
                'type': queueType.value,
                'query': query,
                'lockRequired': lockRequired,
                'function': f,
                "last_run_at": datetime.now()
            }
            if enabled is not None:
                parameters['enabled'] = enabled
            self.queues[name] = parameters
            self.eventQueues[name] = {
                "name": name,
                "function": f,
                "mailbox": [],
                "config": {**DEFAULT_EVENT_CONFIG, **eventConfig},
                "start_time": None  # Time to track, if elapsed and mailbox not empty, trigger event
            }
            return f
        return wrap

    def not_found(self, job_id, update_job, kwargs):
        """This function triggers out if a job fails.

        Args:
            job_id (int): Job ID.
            update_job (method): Job details that will be updated as error.
            kwargs (args): Extra arguments of the job.
        """
        from .Notifications import nextNotifOld
        self.log(LogLevel.ERROR, 'Function not found :(')
        message = 'Funcion no habilitada, favor de contactar a administrador'
        notificationData = {}
        if(nextNotifOld(kwargs['job'])):
            data, errors = self.gql.mutate(mutations.create_notification, {
                'context': 'DANGER',
                'title': f'Error en {kwargs["job"]["queue"]["alias"]}',
                'content': message,
                'metadata': json.dumps({"resolved": False}),
            })
            now = getUtcDate(datetime.now())
            nextNotifDate = now + timedelta(minutes=20)
            notificationData = {
                'id': data['result']['id'],
                'Sent': True,
                'SentAt': now.strftime("%Y-%m-%dT%H:%M:%SZ"),
                # 'frequency': '*/2 * * * *', # cron descriptor or empty string
                # cron descriptor or empty string
                'sendNextAt': nextNotifDate.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }
        update_job({
            'id': job_id,
            'status': 'ERROR',
            'output': json.dumps({
                'notification': notificationData,
                'message': message
            })
        })

    # Run helpers: -------------------------------
    def getJobData(self, job):
        """This function gets the Job data.

        Args:
            job (object): Job that will tried to run.

        Returns:
            Error, job result.
        """
        NOT_FOUND = {
            'alias': 'Funcion no encontrada',
            'description': 'Funcion ejecutada cuando se trata de llamar Job sin una función registrada',
            'schedule': None,
            'enabled': True,
            'type': QueueType.ON_DEMAND,
            'function': self.not_found,
            "last_run_at": datetime.now()
        }
        fn_name = __.get(job, 'queue.name', 'NOT_FOUND')
        job_dict = __.get(self.queues, fn_name, None)
        return fn_name, job_dict

    def groupQueuesByListenStatus(self, queues):
        """This function groups all queues by their listen status.

        Args:
            queues (dict): All queues.

        Returns:
            lists: Ordered lists.
        """
        not_listened_queues = [
            queue['name']
            for queue in queues
            if not __.get(queue, 'listeners', [])
        ]
        worker_listened_queues = [
            queue['name']
            for queue in queues
            if __.find(
                __.get(queue, 'listeners', []),
                lambda listener: (
                    self.worker_code == __.get(listener, 'worker.code') and
                    __.get(listener, 'enabled')
                )
            )
        ]
        worker_disabled_queues = [
            queue['name']
            for queue in queues
            if __.find(
                __.get(queue, 'listeners', []),
                lambda listener: (
                    self.worker_code == __.get(listener, 'worker.code') and
                    not __.get(listener, 'enabled')
                )
            )
        ]
        return (
            sorted(worker_listened_queues),
            sorted(worker_disabled_queues),
            sorted(not_listened_queues),
        )

    def reportListenedQueuesExhaustive(self):
        """This function makes a exhaustive report with logs of the listened queues.
        """
        (
            listened,
            disabled,
            not_listened
        ) = self.groupQueuesByListenStatus(self.currentQueues)
        self.log(LogLevel.SUCCESS, 'LISTENED Queues:')
        for queue_name in listened:
            self.log(LogLevel.SUCCESS, f"\t{queue_name}")
        if not listened:
            self.log(LogLevel.SUCCESS,
                     f'{self.worker_code} not listening to any Queue')
        self.log(LogLevel.WARNING, 'DISABLED Queues:')
        for queue_name in disabled:
            self.log(LogLevel.WARNING, f"\t{queue_name}")
        if not disabled:
            self.log(LogLevel.WARNING, f'\tNo queues disabled')
        self.log(LogLevel.ERROR, 'NOT LISTENED Queues:')
        for queue_name in not_listened:
            self.log(LogLevel.ERROR, f"\t{queue_name}")
        if not not_listened:
            self.log(LogLevel.INFO, f'\tAll queues listened')

    def reportListenedQueuesDiff(self):
        """This function makes a report with logs of the difference of all queues.
        """
        (
            listened,
            disabled,
            not_listened
        ) = self.groupQueuesByListenStatus(self.currentQueues)
        (
            prev_listened,
            prev_disabled,
            prev_not_listened
        ) = self.groupQueuesByListenStatus(self.previousQueues)
        listened_diff = set(listened) - set(prev_listened)
        disabled_diff = set(disabled) - set(prev_disabled)
        not_listened_diff = set(not_listened) - set(prev_not_listened)
        if listened_diff:
            self.log(LogLevel.SUCCESS, f'NEW LISTENED QUEUES:')
            for queue_name in listened_diff:
                self.log(LogLevel.SUCCESS, f'\t{queue_name}')
        if disabled_diff:
            self.log(LogLevel.WARNING, f'NEW DISABLED QUEUES:')
            for queue_name in disabled_diff:
                self.log(LogLevel.WARNING, f'\t{queue_name}')
        if not_listened_diff:
            self.log(LogLevel.ERROR, f'NEW NOT LISTENED QUEUES:')
            for queue_name in not_listened_diff:
                self.log(LogLevel.ERROR, f'\t{queue_name}')

    def reportQueueChanges(self):
        """This function makes a report with all queues changes.
        """
        if not self.previousQueues:
            # first iteration, report exhaustive lists (Listening, skipping, disabled)
            self.reportListenedQueuesExhaustive()
        else:
            # Check for diffs, and only report that
            self.reportListenedQueuesDiff()

    def abortJob(self, job):
        """This function abort a running job.

        Args:
            job (object): Running job.
        """
        variables = {
            'id': job['id'],
            'status': 'ABORTED'
        }
        data, errors = self.gql.mutate(mutations.update_job, variables)

    def lockQueue(self, queue_name):
        """This function locks a queue.

        Args:
            queue_name (string): Queue's name.

        Returns:
            object: Returns lock object.
        """
        variables = {
            'queue': queue_name,
            'worker': self.worker['code'],
            'active': True,
            'startAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
        }
        data, errors = self.gql.mutate(mutations.create_lock, variables)
        lock = __.get(data, 'result')
        queue_index = __.arrays.find_index(
            self.currentQueues, lambda q: q['name'] == queue_name)
        self.currentQueues[queue_index]['locks'].insert(0, lock)
        return lock  # returns lock object

    def unlockQueue(self, queue_name, lock):
        """This function unlocks a queue.

        Args:
            queue_name (string): Queue's name.
            lock (object): Lock object.

        Returns:
            (object): Returns unlocking success status.
        """
        variables = {
            'id': lock['id'],
            'active': False,
            'endAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
        }
        data, errors = self.gql.mutate(mutations.update_lock, variables)
        lock = __.get(data, 'result')
        queue_index = __.arrays.find_index(
            self.currentQueues, lambda q: q['name'] == queue_name)
        self.currentQueues[queue_index]['locks'].insert(0, lock)
        return __.get(data, 'successful')  # returns unlocking success status

    def runJob(self, job):
        """This function runs a job

        Args:
            job (object): Job that will run.

        Raises:
            Exception: Missing status of the job.
        """
        # obtiene lock (if needed):
        lock = None
        if job["queue"]["lockRequired"]:
            lock = self.lockQueue(job["queue"]["name"])
        # Obtiene job
        fn_name, job_dict = self.getJobData(job)
        # ! Validates existence of requested job for this worker:
        if not job_dict:
            self.log(
                LogLevel.ERROR,
                f"{fn_name} job not found (it might exist in another branch / parallel universe), skipping..."
            )
            # TODO: Here we can update the job to aborted or something
            return
        self.log(LogLevel.INFO, f"running {fn_name}")
        job_fn = job_dict['function']
        # Obtiene inputs
        if job['input']:
            try:
                if type(job['input']) == dict:
                    kwargs = job['input']
                else:
                    # FALLBACK, attempt to load as a json-encoded string
                    kwargs = json.loads(job['input'])
            except JSONDecodeError as e:
                # falló al leer datos de input, aborta mision
                variables = {
                    'id': int(job['id']),
                    'status': 'ERROR',
                    'output': f'Job input error: {str(e)}'
                }
                uploaders.update_job(variables)
                self.log(LogLevel.ERROR, e)
                return
        else:
            kwargs = job_dict
        # try to run job

        def update_this_job(variables):
            # * funcion de utilidad para actualizar job sin necesidad de conocer su ID
            status = variables.get('status')
            if not status:
                raise Exception('Missing "status" variable')
            # cast to status type to enforce an enumeration value
            status = JobStatus(status)
            # Extract string from enumeration
            safe_vars = {**variables, 'status': status.value}
            # this allows to specify ID inside job for backwards compatibility
            return uploaders.update_job({'id': job['id'], **safe_vars})

        def update_this_context(context):
            return self.gql.mutate(mutations.update_context, {
                'queue': __.get(job, 'queue.name'),
                'context': json.dumps(context)
            })
        src_status_conn, dst_status_conn = Pipe()

        def get_job_status():
            if dst_status_conn.poll():
                job = dst_status_conn.recv()
                return JobStatus(job['jobStatus'])
            return None
        # ! Inyecta a los kwargs el job que se está ejecutando:
        kwargs['job'] = job
        kwargs['job_id'] = job['id']
        # couple of aliases for the same fn
        kwargs['job']['update'] = update_this_job
        # couple of aliases for the same fn
        kwargs['update_job'] = update_this_job
        kwargs['job']['get_status'] = get_job_status
        kwargs['get_job_status'] = get_job_status
        # acceso directo a atributos del queue al que pertenece el job
        kwargs['queue'] = job['queue']
        kwargs['log'] = self.getProcessLoggerFunction(
            queue=job['queue'], job=job)
        kwargs['context'] = __.get(job, 'queue.context', dict())
        kwargs['update_context'] = update_this_context
        kwargs['kwargs'] = {**kwargs}
        variables = {
            'id': int(job['id']),
            'queueName': job['queue']['name'],
            'status': 'RUNNING',
            'lastRunAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
        }
        data, errors = self.gql.mutate(
            mutation=mutations.run_job, variables=variables)
        job_success = __.get(data, 'updateJob.successful')
        queue_success = __.get(data, 'updateQueue.successful')
        if job_success and queue_success:
            def safe_fn(**kwargs):
                try:
                    job_fn(**kwargs)
                except TypeError as e:
                    self.log(LogLevel.ERROR,
                             f'Error starting job from queue {fn_name}:\n{str(e)}')
                    self.log(LogLevel.ERROR,
                             f'Did you forget to ignore additional params in the job definition?:')
                    self.log(LogLevel.ERROR, f'@vw.job(...)')
                    self.log(
                        LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
                    self.log(LogLevel.ERROR, f'\t...')
            p = self.context.Process(target=safe_fn, kwargs=(kwargs))
            self.log(LogLevel.INFO, f'started job with ID {job["id"]}')
            p.start()
            self.runningJobs[job['id']] = {
                'process': p,
                'job': {**job},
                'status_src': src_status_conn,
                'status_dst': dst_status_conn
            }
            if lock:
                self.lockedJobs.append({
                    'process': p,
                    'job': job,
                    'lock': lock
                })
        else:
            self.log(LogLevel.ERROR, f'Error starting job (queue={fn_name})')

    def queue_locked(self, queue_name):
        """This function checks if a queue is locked.

        Args:
            queue_name (string): Queue's name

        Returns:
            boolean/object: Returns the information of the locked object.
        """
        queues_by_name = __.group_by(self.currentQueues, 'name')
        current_queue = __.get(queues_by_name, f'{queue_name}.0')
        if not __.get(current_queue, 'lockRequired'):
            return False  # Queue does not require locking
        # locked if last lock is active
        return __.get(current_queue, 'locks.0.active')

    def workerQueueEnabled(self, queue):
        """This function checks if the queue is enable.

        Args:
            queue (object): Queue object

        Returns:
            boolean: Returns if the queue is enable.
        """
        # queue not enabled if: no listener for this queue OR current worker not listening OR listening disabled for this worker
        currentQueue = __.find(
            self.currentQueues,
            lambda cq: queue.get('name') and queue['name'] == cq['name']
        )
        listeners = __.get(currentQueue, 'listeners', [])
        if not listeners:
            return False
        listener = __.find(
            listeners,
            lambda lnr: self.worker_code == __.get(lnr, 'worker.code')
        )
        if not listener:
            return False
        if __.get(listener, 'enabled'):
            return True
        return False

    def processJobs(self):
        """This function manages the process of a Job.
        """
        orderBy = self.queryOrderBy.value
        jobs, errors = self.gql.query(queries.all_jobs, variables={
                                      'worker': self.worker['code'],
                                      'orderByJob': {'asc': orderBy}})
        if errors:
            message = errors[0]
            self.log(LogLevel.ERROR, f'Error! msg: {message}')
            return
        if jobs is not None and len(jobs) > 0:
            self.log(LogLevel.WARNING, f"Found {len(jobs)} jobs")
            for job in jobs:
                if self.queue_locked(job['queue']['name']):
                    self.log(LogLevel.ERROR,
                             f"{job['queue']['name']} is LOCKED, job aborted")
                    self.abortJob(job)
                elif self.workerQueueEnabled(job['queue']):
                    self.runJob(job)
                else:
                    self.log(LogLevel.WARNING,
                             f"{job['queue']['name']} is DISABLED, skipping job")
        else:
            self.log(LogLevel.INFO, "Listening to events...")

    def abortStaleJobs(self):
        """This function aborts all stale jobs.
        """
        # * pagination configuration
        workerRegex = f'^{self.worker["code"]}$'  # worker exact match
        limit = 100  # how many jobs to delete in one batch
        cursor = 0  # starting point for each page
        hasNextPage = True
        abortedCount = 0
        pages = 0
        while hasNextPage:
            # * Get staled jobs page
            data, errors = self.gql.query(
                queries.paginate_stale_jobs,
                variables={'worker': workerRegex,
                           'cursor': cursor, 'limit': limit}
            )
            jobs = __.get(data, 'jobs', [])
            cursor = __.get(data, 'pageInfo.endCursor')
            hasNextPage = __.get(data, 'pageInfo.hasNextPage')
            # ! delete 'em jobs
            # * get batch mutation
            BATCH_STALE_DELETE_MUT = mutations.build_batch_stale_mutation(jobs)
            if not BATCH_STALE_DELETE_MUT:
                continue
            # * run mutation
            self.log(LogLevel.WARNING, f'Deleting {len(jobs)} stale jobs...')
            data, errors = self.gql.mutate(BATCH_STALE_DELETE_MUT)
            if errors:
                self.log(LogLevel.ERROR, 'Errors deleting stale jobs:')
                self.log(LogLevel.ERROR, errors)
            self.log(LogLevel.ERROR, f'Deleted {len(jobs)} stale jobs')
            abortedCount += len(jobs)
            pages += 1
        if abortedCount and pages >= 2:
            self.log(LogLevel.ERROR,
                     f'Deleted a total of {abortedCount} stale jobs')
        elif not abortedCount:
            self.log(LogLevel.INFO, 'No stale jobs to delete')

    def updateAvailableQueues(self):
        """This function updates all avaibles queues to the current configuration.
        """
        self.log(LogLevel.SUCCESS, 'Actualizando Jobs disponibles')
        # ! Update each queue / config parameter
        for name, queue in self.queues.items():
            description = __.get(queue, 'description', None)
            schedule = __.get(queue, 'schedule', None)
            query = __.get(queue, 'query', None)
            lockRequired = __.get(queue, 'lockRequired', False)
            variables = {
                'name': queue['name'],
                'alias': queue['alias'],
                'type': queue['type'],
                'lockRequired': lockRequired
            }
            # ! Add optional variables
            if description:
                variables['description'] = description
            if schedule:
                variables['schedule'] = schedule
            if query:
                variables['query'] = query
            # ! Update QUEUE-related config
            data, errors = self.gql.mutate(
                mutation=mutations.upsert_queue,
                variables=variables
            )
            if (errors):
                self.log(LogLevel.ERROR, f'Error actualizando Queue {name}.')
                self.log(LogLevel.ERROR, errors)
            else:
                self.log(LogLevel.SUCCESS,
                         f'Queue {name} actualizado correctamente')

    def updateWorkerQueues(self):
        """This function updates all queues to a current configuration.
        """
        # ! Update WORKER-related config
        # * first, get current configuration (may cause bugs if we create more than one config register, see Many-Many)
        workerQueues, errors = self.gql.query(
            query=queries.get_worker_queues,
            variables={'workerCode': self.worker_code}
        )
        if errors:
            self.log(LogLevel.ERROR, errors)
        # ! Manual Upsert (Not enabled for many-many relationships)
        for name, queue in self.queues.items():
            # ! check whether we should create or update the item:
            workerQueue = __.find(
                workerQueues,
                lambda wk: name == __.get(wk, 'queue.name', None)
            )
            queueEnabled = __.get(queue, 'enabled', False)
            if not workerQueue:
                _, errors = self.gql.mutate(mutations.create_queue_worker, {
                                            'worker': self.worker_code, 'queue': name, 'enabled': queueEnabled})
                if errors:
                    log_lvl = LogLevel.ERROR
                    log_msg = f'Error Creating config for [ {self.worker_code} <=> {name} ]: {errors}'
                else:
                    log_lvl = LogLevel.SUCCESS if queueEnabled else LogLevel.WARNING
                    log_msg = f'Created config for [ {self.worker_code} <=> {name} ] = {"enabled" if queueEnabled else "disabled"}'
            else:
                _, errors = self.gql.mutate(mutations.update_queue_worker, {
                                            'id': workerQueue['id'], 'enabled': queueEnabled})
                if errors:
                    log_lvl = LogLevel.ERROR
                    log_msg = f'Error Updating config for [ {self.worker_code} <=> {name} ]: {errors}'
                else:
                    log_lvl = LogLevel.SUCCESS if queueEnabled else LogLevel.WARNING
                    log_msg = f'Updated config for [ {self.worker_code} <=> {name} ] = {"enabled" if queueEnabled else "disabled"}'
            self.log(log_lvl, log_msg)

    def clearFinishedJobs(self):
        """This function cleans up all finished jobs "cache".
        """
        finished = []
        for job_id, running_job in self.runningJobs.items():
            if not running_job['process'].is_alive():
                finished.append(job_id)
        for f in finished:
            self.runningJobs.pop(f, None)  # remove any finished jobs

    def unlockFinishedJobs(self, force=False):
        """This function updates all finished jobs to unlock them..

        Args:
            force (bool, optional): Option to force the unlock update. Defaults to False.
        """
        unlocked = []
        _log = log if force else self.log
        for i, job_lock in enumerate(self.lockedJobs):
            if not job_lock['process'].is_alive() or force:
                if self.unlockQueue(job_lock['job']['queue']['name'], job_lock['lock']):
                    _log(LogLevel.SUCCESS,
                         f'{job_lock["job"]["queue"]["name"]} UNLOCKED')
                    unlocked.append(i)
                else:
                    _log(LogLevel.ERROR,
                         f'error unlocking {job_lock["job"]["queue"]["name"]}')
        if unlocked:
            unlocked.reverse()
            for u in unlocked:
                self.lockedJobs.pop(u)

    def runTimedOutEvents(self):
        """This function checks all queue that are in time out.
        """
        current_time = time.time()
        for queue in self.eventQueues.values():
            if not queue['start_time'] or len(queue['mailbox']) == 0:
                continue  # nothing to run
            elapsed = current_time - queue['start_time']
            if elapsed >= queue['config']['timeout']:
                self.log(LogLevel.WARNING,
                         f'Event {queue["name"]} timed out, running...')
                self.runEvent(queue['name'])

    def runFrequencyQueues(self):
        """
        This function checks if it has elapsed enough time since the last attempt to run this queue.
        It takes into account the last time a job of this queue was created,
        rather than when was the last time the job RAN, this enforces only 1 attempt per interval.
        """
        from pytz import timezone
        utc = timezone("UTC")
        queuesByType = __.group_by(self.currentQueues, 'type')
        freqQueues = __.get(queuesByType, 'FREQUENCY', default=[])
        # self.log(LogLevel.WARNING, f'checking {len(freqQueues)} frequency queues')
        for queue in freqQueues:
            lastRunAt = __.get(
                queue,
                # ! get from job.updatedAt, rather than queue.lastRunAt itself (see description of this method)
                'jobs.0.updatedAt',
                # ! Default: if it has never been ran, set the lastRunAt as a very old date (forces next run)
                "2000-01-01T00:00:00Z"
            )
            lastRunDate = datetime.strptime(lastRunAt, "%Y-%m-%dT%H:%M:%SZ")
            lastRunDate = utc.localize(lastRunDate)
            now = getUtcDate(datetime.now())
            cron_iter = get_croniter(queue['schedule'], now)
            nextRunAt = [cron_iter.get_next(datetime)
                         for _ in [0, 1]]  # next and nexter dates
            # validate if next date is met:
            elapsedFromLastRun = (
                now - lastRunDate).total_seconds() / 60.0  # in minutes
            freq = (nextRunAt[1] - nextRunAt[0]
                    ).total_seconds() / 60.0  # in minutes
            if (elapsedFromLastRun > freq):
                # ! Run the frequency job!
                if self.workerQueueEnabled(queue):
                    self.gql.mutate(
                        mutation=mutations.create_job,
                        variables={
                            'queueName': queue['name'],
                            'worker': self.worker['code']
                        }
                    )

    def registerJobCreatedSubscription(self):
        """Registers JobCreated subscription that will call certain job when activated"""
        def runOnDemandJob(data):
            """Callback to run when job is created.
            It checks if the job corresponds to active worker or is a dummy subscription call
            """
            job = __.get(data, 'result')
            if not job:
                return  # Possibly a job for another worker
            if self.queue_locked(job['queue']['name']):
                self.log(LogLevel.ERROR,
                         f"{job['queue']['name']} is LOCKED, job aborted")
                self.abortJob(job)
            elif self.workerQueueEnabled(job['queue']):
                self.runJob(job)
            else:
                self.log(LogLevel.WARNING,
                         f"{job['queue']['name']} is DISABLED, skipping job")

        self.unsubJobCreated = self.gql.subscribe(
            query=subscriptions.JOB_CREATED,
            variables={'workerCode': self.worker['code']},
            callback=runOnDemandJob)
        if (self.unsubJobCreated):
            self.log(LogLevel.SUCCESS, 'SUBSCRIBED TO JOB CREATIONS')
        else:
            self.log(LogLevel.ERROR, 'ERROR SUBSCRIBING TO JOB CREATIONS')

    def runEvent(self, queue_name):
        '''
        This wrapper ensures that whenever runEvent is called, the event is cleared
        '''
        cleanup = self.actuallyRunEvent(queue_name)
        # Move mailbox data and clear it
        if cleanup:
            self.eventQueues[queue_name]["mailbox"] = []
            self.eventQueues[queue_name]["start_time"] = None  # Clear timeout

    def actuallyRunEvent(self, queue_name):
        """Runs job of type ON_EVENT

        Parameters
        ----------
        queue_name : string
          Name of the queue to run
        Returns
        -------
        boolean
          True if a cleanup should be performed (event cleanup)
          False if cleanup should be skipped (to run event at a later point)
        """
        queuesByName = __.group_by(self.currentQueues, 'name')
        currentQueue = __.get(queuesByName, f'{queue_name}.0')
        if self.queue_locked(currentQueue['name']):
            self.log(LogLevel.WARNING,
                     f"{currentQueue['name']} is LOCKED, event postponed")
            return False
        elif not self.workerQueueEnabled(currentQueue):
            self.log(
                LogLevel.WARNING, f'Event triggered for {queue_name} but Queue not enabled, skipping...')
            return True
        data = self.eventQueues[queue_name]["mailbox"]
        # ! try to run job
        # obtiene lock (if needed):
        lock = None
        if currentQueue["lockRequired"]:
            lock = self.lockQueue(currentQueue["name"])
        # Crea job asociado a evento
        response, errors = self.gql.mutate(
            mutation=mutations.run_event_job,
            variables={
                'queueName': queue_name,
                'worker': self.worker['code'],
                'lastRunAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
            }
        )
        if errors:
            self.log(LogLevel.ERROR, f'Error starting event:\n{errors}')
            return True
        job = __.get(response, 'job.result')

        def update_this_job(variables):
            # * funcion de utilidad para actualizar job sin necesidad de conocer su ID
            status = variables.get('status')
            if not status:
                raise Exception('Missing "status" variable')
            # cast to status type to enforce an enumeration value
            status = JobStatus(status)
            # Extract string from enumeration
            safe_vars = {**variables, 'status': status.value}
            # this allows to specify ID inside job for backwards compatibility
            return uploaders.update_job({'id': job['id'], **safe_vars})

        def update_this_context(context):
            return self.gql.mutate(mutations.update_context, {
                'queue': __.get(currentQueue, 'name'),
                'context': json.dumps(context)
            })
        src_status_conn, dst_status_conn = Pipe()

        def get_job_status():
            if dst_status_conn.poll():
                job = dst_status_conn.recv()
                return JobStatus(job['jobStatus'])
            return None
        # ! Inyecta a los kwargs el job que se está ejecutando:
        kwargs = {
            'job_id': job['id'],
            'job': {**job, 'update': update_this_job, 'get_status': get_job_status},
            'update_job': update_this_job,
            'get_job_status': get_job_status,
            'queue': currentQueue,
            'context': __.get(currentQueue, 'context', dict()),
            'update_context': update_this_context,
            'data': data,
            'log': self.getProcessLoggerFunction(queue=currentQueue, job=job)
        }
        kwargs['kwargs'] = {**kwargs}

        def safe_fn(**kwargs):
            try:
                self.eventQueues[queue_name]['function'](**kwargs)
            except TypeError as e:
                self.log(
                    LogLevel.ERROR, f'Error starting job from queue {queue_name}:\n{str(e)}')
                self.log(
                    LogLevel.ERROR, f'Did you forget to ignore additional params in the job definition?:')
                self.log(LogLevel.ERROR, f'@vw.job(...)')
                self.log(
                    LogLevel.ERROR, f'def my_job(my_param, my_param2, **_) # <== can be **kwargs or **_')
                self.log(LogLevel.ERROR, f'\t...')
        p = self.context.Process(target=safe_fn, kwargs=(kwargs))
        p.start()
        self.runningJobs[job['id']] = {
            'process': p,
            'job': {**job},
            'status_src': src_status_conn,
            'status_dst': dst_status_conn
        }
        if lock:
            self.lockedJobs.append({
                'process': p,
                'job': {**job, 'queue': {**currentQueue}},
                'lock': lock
            })
        return True

    def getOnQueueEventCallback(self, queue: Dict[str, str]) -> Callable:
        """Wrap the onEventTriggered with a given queue data estructure for labeling purposes

        Args:
            queue (Dict[str, str]): Data structure returned by self.queues
        """
        def onEventTriggered(data):
            """Callback to execute on every event triggered of the same type.
            This callback performs a check to know if enough data (or time)
            has elapsed and thus the Job has to be executed with the current chunk of data

            Parameters
            ----------
            data : any
                Data received from the event
            """
            if not self.workerQueueEnabled(queue):
                return  # Skipping event
            if len(self.eventQueues[queue["name"]]["mailbox"]) == 0:
                self.log(
                    LogLevel.WARNING, f'EVENT { queue["name"] } TRIGGERED, BUILDING BATCH')
                self.eventQueues[queue["name"]]["start_time"] = time.time()
            self.eventQueues[queue["name"]]["mailbox"].append(data)
            if len(self.eventQueues[queue["name"]]["mailbox"]) >= self.eventQueues[queue["name"]]['config'][
                    'max_batch']:
                """Run event only if event surpassed the upper data limit"""
                self.runEvent(queue["name"])
        return onEventTriggered

    def registerEventJobs(self):
        """This function registers a new event job.
        """
        queuesByType = __.group_by(self.queues.values(), 'type')
        eventQueues = __.get(queuesByType, 'EVENT', default=[])
        for queue in eventQueues:
            unsub = self.gql.subscribe(
                query=queue['query'],
                callback=self.getOnQueueEventCallback(queue))
            if (unsub):
                self.log(
                    LogLevel.SUCCESS, f'REGISTERED LISTENER FOR EVENT QUEUE { queue["name"] }')
            else:
                self.log(
                    LogLevel.ERROR, f'ERROR REGISTERING LISTENER FOR EVENT QUEUE { queue["name"] }')

    def registerJobUpdatesListener(self):
        """This function registers for all jobs updates.
        """
        def onJobUpdated(data):
            successful = __.get(data, 'successful')
            job = __.get(data, 'result')
            # check if update even succeeded
            if not job or not successful:
                return
            # check if updated job matches any currently running
            job_running = self.runningJobs.get(job['id'], None)
            if not job_running:
                return
            # at this point we are certain the updated job is running, update structure to inform the job:
            status_conn = self.runningJobs[job['id']]['status_src']
            status_conn.send(job)
        unsub = self.gql.subscribe(
            query=subscriptions.JOB_UPDATED, callback=onJobUpdated)
        if (unsub):
            self.log(LogLevel.SUCCESS, f'REGISTERED LISTENER FOR JOB UPDATES')
        else:
            self.log(LogLevel.ERROR,
                     f'ERROR REGISTERING LISTENER FOR JOB UPDATES')

    def registerContinuousJobs(self):
        """This function registers for all continues jobs.
        """
        queuesByType = __.group_by(self.queues.values(), 'type')
        shedeuleQueues = __.get(queuesByType, 'CONTINUOUS', default=[])
        for shedeuleQueue in shedeuleQueues:
            queue_name = __.get(shedeuleQueue, 'name', None)
            if queue_name:
                self.continuousJobs.append(queue_name)
                orderBy = self.queryOrderBy.value
                queueLocks, errors = self.gql.query(queries.get_queueLocks, variables={
                    "workerCode": self.worker_code,
                    "queueName": queue_name,
                    'orderByQueueLock': {'desc': orderBy}})
                if errors:
                    self.log(LogLevel.ERROR,
                             f'Error get_queueLocks:\n{errors}')
                    continue
                if queueLocks:
                    self.log(
                        LogLevel.WARNING, f'{queue_name} queue (CONTINUOUS type) has queueLocks')
                    for queueLock in queueLocks:
                        variables = {
                            'id': __.get(queueLock, 'id', None),
                            'active': False,
                            'endAt': getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
                        }
                        _, errors = self.gql.mutate(
                            mutations.update_lock, variables)
                        if errors:
                            self.log(LogLevel.ERROR,
                                     f'Error update_lock:\n{errors}')

        if self.continuousJobs:
            self.log(LogLevel.SUCCESS, 'REGISTERED Continuous Queues:')
            for continuousJob in self.continuousJobs:
                self.log(LogLevel.SUCCESS, f"\t{continuousJob}")

    def monitorContinuousJobs(self):
        """This function checks all continous jobs.
        """
        running_jobs_names = [__.get(running_job, 'job.queue.name', None)
                              for _, running_job in self.runningJobs.items()]
        for continuousJob in self.continuousJobs:
            if not continuousJob in running_jobs_names and self.queues[continuousJob]['enabled']:
                self.log(LogLevel.WARNING,
                         f'Continuous Job {continuousJob} is not running')
                self.log(LogLevel.INFO, f'Creating Job {continuousJob} ...')
                _, errors = self.gql.mutate(mutation=mutations.create_job,
                                            variables={'queueName': continuousJob, 'worker': self.worker['code']})
                if errors:
                    self.log(LogLevel.ERROR, f'Error create_job:\n{errors}')

    def getProcessLoggerFunction(self, queue=None, job=None):
        """Logs to STDOUT formatted to help get helpful information

        Parameters
        ----------
        queue : int, optional
          ID for the queue sending this message, by default 0
        job : int, optional
          ID for the job sending this message, by default 0
        """
        job_zero_padding = len(str(__.get(self.worker, 'jobs.0.id')))
        # no need to pass as argument, will always be the same
        W = str(self.worker["id"]).zfill(2)
        Q = str(__.get(queue, 'id', 0)).zfill(3)
        J = str(__.get(job, 'id', 0)).zfill(job_zero_padding)
        log_queue_key = __.get(queue, 'name', 'main')
        log_queue = self.logQueues[log_queue_key]
        if self.loggingStyle == LogStyle.PREFIX_ALL_LINES:
            def sendLogToQueue(level, message):
                D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
                msgs = message.split('\n')
                for msg in msgs:
                    log_queue.put({
                        'level': level,
                        'content': f'[D={D}/W={W}/Q={Q}/J={J}] - {msg}'
                    })
        else:
            def sendLogToQueue(level, message):
                D = getUtcDate(datetime.now()).strftime("%Y-%m-%dT%H:%M:%SZ")
                log_queue.put({
                    'level': level,
                    'content': f'[D={D}/W={W}/Q={Q}/J={J}] - {message}'
                })
        return sendLogToQueue

    def startLoggingLoop(self):
        """This function has the responsability to make a log loop for any event.
        """
        self.log = self.getProcessLoggerFunction()
        # FN DEFINITION --------------------------

        def loggingLoop():
            try:
                while not self.stopLogging:
                    # * check mailbox for every queue
                    # * log as many messages from same queue as possible
                    # TODO: May set a limit to this, in message count or timeout for other queues)
                    # ! Send messages to log-upload queue
                    # ! flag as CREATE or UPSERT for retention or real-time visualization only
                    for _, log_queue in self.logQueues.items():
                        while not log_queue.empty():
                            msg = log_queue.get()
                            log(msg['level'], msg['content'])
            except KeyboardInterrupt:
                log(LogLevel.WARNING, 'Logging stopped due to keyboard interrupt')
            log(LogLevel.WARNING, 'Logging loop stopped')
        # END FN DEFINITION --------------------------
        # Create Queues for each queue:
        for queue_name, _ in self.queues.items():
            self.logQueues[queue_name] = SimpleQueue()
        self.loggingProcess = self.context.Process(target=loggingLoop)
        self.loggingProcess.start()
        self.log(LogLevel.SUCCESS, 'Logging loop started')

    def stopLoggingLoop(self):
        """This function stops de logging loop.
        """
        self.stopLogging = True
        self.loggingProcess.join()

    def eventLoop(self, interval=1.0):
        """This function manages all generic worker activities.

        Args:
            interval (float, optional): Interval time of each loop. Defaults to 1.0.
        """
        try:
            orderBy = self.queryOrderBy.value
            currentQueues, errors = self.gql.query(
                queries.all_queues_w_listeners,
                variables={
                    'orderByJob': {'desc': orderBy},
                    'orderByQueueLock': {'desc': orderBy}})
            if (errors):
                self.log(LogLevel.ERROR, f'errors: {errors}')
                return
            self.currentQueues = currentQueues
            self.reportQueueChanges()
            self.clearFinishedJobs()
            self.unlockFinishedJobs()
            self.runFrequencyQueues()
            self.runTimedOutEvents()
            # TODO: sendReminders(queues)
            if self.pollingMode == PollingMode.QUERY:
                self.processJobs()
            else:
                self.log(LogLevel.INFO, 'Listening to events...')
            if self.useRedis:
                self.redis.monitorRedisSubscriptions()
            self.monitorContinuousJobs()
        except Exception as e:
            self.log(LogLevel.ERROR, f"Error in worker's event loop!")
            self.log(LogLevel.ERROR, f'e: {e}')
        self.previousQueues = self.currentQueues
        time.sleep(interval)

    def getWorker(self):
        """This function gets the worker information.
        """
        orderBy = self.queryOrderBy.value
        worker, errors = self.gql.query_one(
            queries.get_worker,
            {'code': self.worker_code, 'orderByJob': {'desc': orderBy}})
        if errors:
            log(LogLevel.ERROR, errors)
            raise Exception('Worker not found')
        self.worker = worker

    def run(self, interval=1.0):
        """This function turns on the worker.

        Args:
            interval (float, optional): Time interval of each eventloop.
              Defaults to 1.0.
        """
        self.context = mp.get_context('fork')
        self.getWorker()
        self.startLoggingLoop()
        self.abortStaleJobs()
        if (self.jobConfigMode == JobConfigMode.SYNC):
            self.updateAvailableQueues()
            self.updateWorkerQueues()
        else:
            self.log(
                LogLevel.INFO, 'No config sync performed. If required, set jobConfigMode to SYNC')
        if (self.pollingMode == PollingMode.SUBSCRIPTION):
            self.registerJobCreatedSubscription()
        self.registerEventJobs()
        self.registerJobUpdatesListener()
        if self.useRedis:
            self.redis.setLog(self.log)
            self.redis.initializeRedis()
        self.registerContinuousJobs()
        while 1:
            try:
                self.eventLoop(interval=interval)
            except KeyboardInterrupt:
                log(LogLevel.WARNING, "\nuser's stop signal, exiting...")
                self.stopLoggingLoop()
                self.unlockFinishedJobs(force=True)
                self.gql.close()
                break




JOB_CREATED = '''
subscription  JobCreated(
  $workerCode: String!
){
  jobCreated(filter: { workerCode: $workerCode }){
    successful
    messages{field message}
    result{
      id
      userId
      input
      context
      output
      progress
      jobStatus
      queue {
        id
        name
        alias
        type
        enabled
        context # for rehydration
        schedule
        query
        lockRequired
        # locks(orderBy:{desc:ID} limit: 1){
          # id
          # active
          # startAt
          # endAt
        # }
      }
      worker{
        id
        code
        name
      }
    }
  }
}
'''

JOB_UPDATED = '''
subscription {
  jobUpdated{
    successful
    messages{
      field
      message
    }
    result{
      id
      jobStatus
    }
  }
}
'''


import pydash as __
import time
from .Logging import LogLevel
from pygqlc import GraphQLClient
from multiprocessing import Event
import redis

DEFAULT_REDIS_CONFIG = {
  "host": 'localhost', 
  "port": 6379,
  "db": 0,
  "url": ""
}

class WorkerRedis():
    def __init__(self):
        self.redisConfig = DEFAULT_REDIS_CONFIG
        self.redis = None
        self.connected = Event()
        self.connected.clear()
        self.startTime = None
        self.redisSubscriptions = []
        self.runningSubscriptions = []
        self.log = None
        self.gql = GraphQLClient()

    def setRedisConfig(self, host='localhost', port=6379, db=0, url=""):
        REDIS_CONFIG = {
            "host": host, 
            "port": port,
            "db": db,
            "url": url
        }
        self.redisConfig = REDIS_CONFIG

    def setRedisConnection(self):
        if self.redisConfig['url']:
            self.redis = redis.Redis.from_url(self.redisConfig['url'])
        else:
            self.redis = redis.Redis(host = self.redisConfig['host'], 
                            port=self.redisConfig['port'], 
                            db=self.redisConfig['db'])

    def setLog(self, log):
        self.log = log

    def getClient(self):
        return self.redis

    def isConnected(self):
        return self.connected.is_set()
        
    def setRedisSubscriptions(self, name=None, subscription=None, key=None, value=None, prefix=None, suffix=None, reconnectionTime=600):
        if name and subscription and key and value:
            sub = {'name': name, 
                   'subscription': subscription, 
                   'key': key, 
                   'value': value,
                   'prefix': prefix,
                   'suffix': suffix,
                   'reconnectionTime': reconnectionTime}
            self.redisSubscriptions.append(sub)

    def createCallbackRedisFunction(self, name, dkey, dvalue, prefix, suffix, expireTime):
        def callbackRedisFunction(message):
            key = __.get(message, dkey, None)
            value = __.get(message, dvalue, None)
            
            if key and (value is not None):
                if prefix:
                    key = prefix + '_' + key
                if suffix:
                    key = key + '_' + suffix
                self.redis.set(key, value)
                self.redis.publish(key, value)
                self.redis.set(name, 'Alive')
                self.redis.expire(name, expireTime)
            else: 
                self.log(LogLevel.ERROR, f'Not {dkey} or {dvalue} in {name} subscription')
        return callbackRedisFunction

    
    def initializeSubscription(self, redisSub):
        callbackRedisFunction = self.createCallbackRedisFunction(redisSub['name'], redisSub['key'], redisSub['value'],
                                                                 redisSub['prefix'], redisSub['suffix'], redisSub['reconnectionTime'])
        unsub = self.gql.subscribe(query=redisSub['subscription'], callback=callbackRedisFunction)
        name =  redisSub['name']
        if (unsub):
            self.log(LogLevel.SUCCESS, f'Initialize Redis subscription: {name}')
            runningSup = {'redisSub': redisSub, 'unsub': unsub}
            self.runningSubscriptions.append(runningSup)
            self.redis.set(redisSub['name'], 'Alive')
            self.redis.expire(redisSub['name'], redisSub['reconnectionTime'])
        else:
            self.log(LogLevel.ERROR, f'ERROR Initialize Redis subscription: {name}')


    def initializeRedis(self):
        self.log(LogLevel.INFO, 'Worker using Redis service')
        self.setRedisConnection()
        try:
            ping = self.redis.ping()
            self.log(LogLevel.SUCCESS, f'Redis ping - {ping}')
        except:
            self.log(LogLevel.ERROR, 'Redis connection error, check connection parameters')
            self.startTime = time.time()
            self.connected.clear()
            return
    
        self.redis.flushall()  

        for redisSub in self.redisSubscriptions:
            self.initializeSubscription(redisSub)
        self.connected.set()
        self.log(LogLevel.SUCCESS, 'Redis Worker is connected')
        
    def monitorRedisSubscriptions(self):
        if not self.isConnected():
            self.waitingRedisConnection()
            return
        
        for idx, sub in enumerate(self.runningSubscriptions):
            isVarRedis = self.redis.exists(sub['redisSub']['name'])
            name = sub['redisSub']['name']
            if not self.redis.exists(sub['redisSub']['name']):
                name = sub['redisSub']['name']
                self.log(LogLevel.WARNING, f'Redis subscription {name} expired time')
                self.gql.resetSubsConnection()
                self.redis.set(name, 'Alive')
                self.redis.expire(name, sub['redisSub']['reconnectionTime'])
                self.log(LogLevel.INFO, f'Connecting Redis subscription: {name} ...')

    def waitingRedisConnection(self):
        waitingTime = time.time() - self.startTime
        if waitingTime > 30:
            self.initializeRedis()

            



from pprint import pprint
from enum import Enum
from datetime import datetime
from pytz import timezone
import pydash
from pygqlc import GraphQLClient
from . import queries
from . import mutations
from .Logging import log, LogLevel
from .mutations import update_notification_metadata
from .dateHelpers import getUtcDate
from .croniterHelpers import get_croniter
gql = GraphQLClient()
# * for correct time comparisons between server and client
utc = timezone('UTC')


class NotificationBehaviour(Enum):
    DEFAULT = 'DEFAULT'
    PERSISTENT = 'PERSISTENT'


class NotificationType(Enum):
    PRIMARY = 'PRIMARY'
    SECONDARY = 'SECONDARY'
    INFO = 'INFO'
    SUCCESS = 'SUCCESS'
    WARNING = 'WARNING'
    DANGER = 'DANGER'


reminders = {}
reminderCheckers = {}


def reminder(name):
    def wrap(f):
        reminders[name] = f  # ! function that will get the reminder content
        return f
    return wrap


def reminderChecker(name):
    def wrap(f):
        # ! function that will eval if a notification must be sent again
        reminderCheckers[name] = f
        return f
    return wrap


def defaultReminderChecker(schedule, now, lastSentDate):
    log(LogLevel.DEBUG, 'Running default reminder checker')
    log(LogLevel.DEBUG, schedule)
    log(LogLevel.DEBUG, now)
    log(LogLevel.DEBUG, lastSentDate)
    cron_iter = get_croniter(schedule, now)
    nextRunAt = [cron_iter.get_next(datetime)
                 for _ in [0, 1]]  # next and nexter dates
    # validate if next date is met:
    elapsedFromLastRun = (
        now - lastSentDate).total_seconds() / 60.0  # in minutes
    freq = (nextRunAt[1] - nextRunAt[0]).total_seconds() / 60.0  # in minutes
    if (elapsedFromLastRun > freq):
        log(LogLevel.DEBUG, 'Return true: Frequency for next notification met')
        return True
    log(LogLevel.DEBUG, 'Return false: Frequency not met')
    return False  # Frequency not met


def defaultReminder(queue):
    return {
        'title': 'Recordatorio',
        'content': f'Hay una notificación sin resolver para {queue["alias"]}',
        'context': NotificationType.INFO.value,
        'link': 'monitors',
        'linkText': 'Ir a Monitores'
    }


def resolve(queue):
    import json
    # * check if there is any notification to resolve
    notificationId = pydash.get(queue, 'query.notification.id', None)
    if (notificationId):
        # * if so, resolve notification
        metadata = json.dumps({'resolved': True})
        response, errors = gql.mutate(
            mutation=update_notification_metadata,
            variables={
                'id': parseInt(notificationId),
                'metadata': metadata
            }
        )
        if (errors):
            log(LogLevel.ERROR, 'Error resolving notification!:')
            log(LogLevel.ERROR, errors)


# ! Validate if a new notification must be sent
'''
Job with notification output example:
{
  "id": "590",
  "queue": {
    "id": "17",
    "name": "BLABLA_TEST"
  },
  "progress": 0,
  "jobStatus": "ERROR",
  "input": "{}",
  "output": {
    "notification": {
      "id": "123"
      "Sent": true, 
      "SentAt": "2019-07-26T12:51:16Z", 
      "frequency": ", # cron descriptor or empty string
      "sendNextAt": "2019-07-26T12:51:16Z"
    },
    "message": "Funcion no habilitada, favor de contactar a administrador"},
  "insertedAt": "2019-07-26T12:51:14Z"
}
'''
# ! deprecated, for reference only


def nextNotifOld(job, worker=None):
    log(LogLevel.DEBUG, 'Debugging nextNotification')
    queueName = pydash.get(job, 'queue.name')
    queue, errors = gql.query(queries.get_queue_last_jobs, {
        'name': queueName,
        'count': 2,
        'worker': worker
    })
    prev_job_output = pydash.get(queue, 'jobs.1.output')
    try:
        prev_job_output = json.loads(prev_job_output)
    except Exception as e:
        prev_job_output = {}
    notif_data = pydash.get(prev_job_output, 'notification')
    if not prev_job_output or not notif_data:
        log(LogLevel.DEBUG, 'Return true: No notification info provided')
        return True  # * if no notification info is provided, the validation passes
    # * If Notification info is provided, get notification
    log(LogLevel.DEBUG, notif_data)
    notification, errors = gql.query(queries.get_notification, {
                                     'id': int(notif_data['id'])})
    if (errors):
        log(LogLevel.ERROR, f'Error al leer notificacion: {errors}')
        return True
    # * If the notification is marked as resolved, you can send notifications again
    metadata = pydash.get(notification, 'metadata', "{}")
    log(LogLevel.DEBUG, f'metadata: {metadata}')
    try:
        metadata = json.loads(metadata)
    except Exception as e:
        metadata = {}
    resolved = pydash.get(metadata, 'resolved', True)
    if resolved:
        log(LogLevel.DEBUG, 'Return true: Previous notification resolved already')
        return True
    # * If not resolved (or not resolvable), check if the time has come to send a new notification
    now = getUtcDate(datetime.now())
    sendNextAt = pydash.get(notif_data, 'sendNextAt')
    log(LogLevel.DEBUG, f'sendNextAt: {sendNextAt}')
    if (sendNextAt):
        sendNextDate = datetime.strptime(sendNextAt, "%Y-%m-%dT%H:%M:%SZ")
        sendNextDate = utc.localize(sendNextDate)
        if now > sendNextDate:
            log(LogLevel.DEBUG, 'Return true: Next notification date Met')
            return True
        else:
            log(LogLevel.DEBUG, 'Return false: Next notification date Not yet met')
            return False
    frequency = pydash.get(notif_data, 'frequency')
    if (frequency):
        sentAt = pydash.get(notif_data, 'sentAt')
        sentAtDate = datetime.strptime(sentAt, "%Y-%m-%dT%H:%M:%SZ")
        sentAtDate = utc.localize(sentAtDate)
        croniter = get_croniter(frequency, now)
        nextRunAt = [cron_iter.get_next(datetime)
                     for _ in [0, 1]]  # next and nexter dates
        # validate if next date is met:
        elapsedFromLastRun = (
            now - sentAtDate).total_seconds() / 60.0  # in minutes
        freq = (nextRunAt[1] - nextRunAt[0]
                ).total_seconds() / 60.0  # in minutes
        if (elapsedFromLastRun > freq):
            log(LogLevel.DEBUG, 'Return true: Frequency for next notification met')
            return True
        log(LogLevel.DEBUG, 'Return false: Frequency not met')
        return False  # Frequency not met
    log(LogLevel.ERROR, 'Weird notification case, should not happen :(')
    log(LogLevel.ERROR, prev_job_output)
    log(LogLevel.ERROR, notification)
    return False


def validateNextNotification(queue):
    import json
    # ! Get required data and Clean data validations:
    checker = pydash.get(reminderCheckers, queue['name'], None)
    if not reminderChecker:
        log(LogLevel.ERROR, "No es posible enviar notificaciones inteligentes sin configurar una funcion @Notifications.reminder")
        return {'ready': False}
    queueConfigString = pydash.get(queue, 'query', "{}")
    if not queueConfigString:
        log(LogLevel.DEBUG, "With no queue configuration data, is assumed to be ok to send a new notification")
        return {'ready': True, 'isReminder': False}
    queueConfig = json.loads(queueConfigString)
    notificationData = pydash.get(queueConfig, 'notification', {})
    if not notificationData:
        log(LogLevel.DEBUG, "With no notification data, is assumed to be ok to send a new notification")
        return {'ready': True, 'isReminder': False}
    # * if notification data available, get it and check for resolvedness:
    lastNotification, errors = gql.query(queries.get_notification, {
                                         'id': int(notificationData['id'])})
    if (errors):
        log(LogLevel.ERROR, f'Error al leer notificacion: {errors}')
        return {'ready': False}
    # * If the notification is marked as resolved, you can send notifications again
    metadata = pydash.get(lastNotification, 'metadata', "{}")
    try:
        metadata = json.loads(metadata)
    except Exception as e:
        metadata = {}
    resolved = pydash.get(metadata, 'resolved', True)
    if resolved:
        log(LogLevel.DEBUG, 'Return true: Previous notification resolved already')
        return {'ready': True, 'isReminder': False}
    log(LogLevel.DEBUG, 'Notification not resolved yet, evaluating reminderChecker')
    # * First, we get the two possible dates:
    lastReminderAt = pydash.get(queueConfig, 'reminder.insertedAt', None)
    lastNotificationAt = pydash.get(lastNotification, 'insertedAt', None)
    # ! if a reminder was sent recently, pick it over the last notification
    lastSentAt = lastReminderAt if lastReminderAt else lastNotificationAt
    if lastReminderAt:
        log(LogLevel.DEBUG,
            f'reminder has been sent, last time was: {lastReminderAt}')
    else:
        log(LogLevel.DEBUG,
            f'reminder has not been sent, last notification was at: {lastNotificationAt}')
    # get dates for comparisons
    now = getUtcDate(datetime.now())
    lastSentDate = datetime.strptime(lastSentAt, "%Y-%m-%dT%H:%M:%SZ")
    lastSentDate = utc.localize(lastSentDate)
    mustRemind = checker(now, lastSentDate)
    return {'ready': mustRemind, 'isReminder': True}


def smartNotification(queue, content):
    """This function makes an action notification.

    Args:
        queue (str): Action that the notification do.
        content (dic): Variables that complete the message of the notification.
    """
    import json
    queueName = pydash.get(queue, 'name')
    # ! Validate if new notification must be sent
    nextNotification = validateNextNotification(queue)
    if not nextNotification['ready']:
        return
    isReminder = nextNotification['isReminder']
    # ! injects the resolved field to metadata only when this is the first notification
    metadata = json.dumps({'resolved': False} if not isReminder else {})
    variables = {}
    if isReminder:
        _reminder = reminders[queueName](queue)
        variables = {
            **_reminder,
            'metadata': metadata
        }
    else:
        variables = {
            **content,
            'metadata': metadata
        }
    response, errors = gql.mutate(mutations.create_notification, variables)
    if errors:
        log(LogLevel.ERROR, 'Error sending smart notification:')
        log(LogLevel.ERROR, errors)
        return
    log(LogLevel.SUCCESS, 'Notification sent successfully:')
    notification = pydash.get(response, 'result')
    # * get queue metadata and inject notification data:
    queue_meta_string = pydash.get(queue, 'query', "{}")
    queue_metadata = json.loads(
        queue_meta_string if queue_meta_string else "{}")
    if isReminder:
        queue_metadata = {
            **queue_metadata,
            'reminder': notification,
        }
    else:
        queue_metadata = {
            **queue_metadata,
            'notification': notification,
        }
    variables = {
        'queueName': queue['name'],
        # * get all notification params
        'metadata': json.dumps(queue_metadata),
    }
    response, errors = gql.mutate(mutations.update_queue_meta, variables)
    if errors:
        log(LogLevel.ERROR, 'Error saving smart notification data (to Queue):')
        log(LogLevel.ERROR, errors)
        return


def notification(content):
    """This function makes a simple notification,

    Args:
        content (dict): Variables that complete the message of the notification

    Returns:
        GraphqlResponse: Returns a GraphqlResponse, a succesful message.
    """
    response, errors = gql.mutate(
        mutations.create_notification, variables=content)
    return response, errors


def groupNotification(
    groupName,
    content=None,  # dictionary containing the static information to display in the Notification
    # TODO: function that gets the user info as input and returns the content as output
    contentBuilder=None
):
    """This function makes a simple notification to a group of "users". 

    Args:
        groupName (str): Group's name
        content (dict, optional): Variables that complete the message of the 
          notification. Defaults to None.

    Returns:
        dict: Returns a succesful message.
    """
    # ! get groups information
    groupUsers, errors = gql.query(queries.get_group_users, variables={
                                   'groupName': groupName})
    if errors:
        return {'successful': False, 'errors': errors}
    if len(groupUsers) == 0:
        return {'successful': False, 'errors': ['No group users for the selected group.']}
    # ! if group exists, send notifications for the entire group:
    all_errors = []
    for groupUser in groupUsers:
        userId = pydash.get(groupUser, 'user.id')
        _, errors = gql.mutate(
            mutations.create_notification,
            variables={
                **content,
                # Attaches the user information to the content of the notification
                'userId': int(userId)
            }
        )
        all_errors.extend(errors)
    # ! if something wrong, inform the error(s) to the caller function:
    if all_errors:
        return {'successful': False, 'errors': all_errors}
    # ! if everything ok, notify the caller function the OK output:
    else:
        return {'successful': True, 'errors': []}



get_worker='''
query GetWorker($code: String!, $orderByJob: OrderByJob) {
  worker(findBy: { code: $code }) {
    id
    code
    name
    jobs(orderBy: $orderByJob, limit: 1) {
      id
    }
  }
}
'''

all_jobs='''
query WorkerJobs($worker: String, $orderByJob: OrderByJob){
    jobs(
        orderBy: $orderByJob, 
        filter: {
          jobStatus: WAITING
          workerCode: $worker
        }
    ) {
        id
        userId
        input
        context
        output
        progress
        jobStatus
        queue {
          id
          name
          alias
          type
          context # for rehydration
          enabled
          schedule
          query
          lockRequired
          # locks(orderBy:{desc:ID} limit: 1){
            # id
            # active
            # startAt
            # endAt
          # }
        }
    }
}
'''

waiting_jobs = '''
query WorkerWaitingJobs($worker: String){
  jobs(filter: {
    workerCode: $worker
    jobStatus: WAITING
  }) {
    id
  }
}
'''

paginate_stale_jobs = '''
  query WorkerWaitingJobsPaginate(
    $worker: String!
    $cursor: Int = 0
    $limit: Int = 10
  ){
    jobsPaginate(
      cursor: $cursor
      filter: {
        workerCode: $worker
        jobStatus: WAITING
      }
      orderBy: {asc: ID}
      limit: $limit
    ) {
      pageInfo { endCursor hasNextPage }
      jobs: edge{ id }
    }
  }
'''

get_job ='''
query GetJob(
  $id: ID!
){
    job(
        id: $id
    ) {
        id
        input
        output
        jobStatus
        queue {
          id
          name
          alias
          type
          enabled
          schedule
          query
        }
    }
}
'''

get_queue = '''
query QueueByName(
  $queueName: String
)
{
  queue(
    findBy:
    {name: $queueName}
  ){
    id
    name
    alias
    type
    enabled
    schedule
    query
  }
}
'''

all_queues_w_listeners = '''
query($orderByJob: OrderByJob, $orderByQueueLock: OrderByQueueLock) {
  queues {
    id
    name
    alias
    description
    context
    type
    schedule
    query
    lastRunAt
    lockRequired
    jobs(orderBy: $orderByJob, limit: 1) {
      id
      updatedAt
    }
    locks(orderBy: $orderByQueueLock, limit: 1) {
      id
      active
      startAt
      endAt
    }
    listeners: queueWorkers {
      id
      enabled
      worker {
        id
        code
      }
    }
    updatedAt
    insertedAt
  }
}
'''

get_queue_last_jobs = '''
query QueueLastJobs(
  $name: String!
  $count: Int!
  $worker: String
){
  queue(
    findBy: {
      name:$name
    }
  ){
    id
    name
    alias
    enabled
    jobs(
      filter:{workerCode: $worker}
      orderBy:
      {desc: INSERTED_AT}
      limit: $count
    ){
      id
      jobStatus
      output
    }
  }
}
'''

get_queue_last_job = '''
query QueueLastJob(
  $name: String!
  $worker: String
){
  queue(
    findBy: {
      name:$name
    }
  ){
    id
    name
    alias
    enabled
    jobs(
      filter:{workerCode: $worker}
      orderBy:
      {desc: INSERTED_AT}
      limit: 1
    ){
      id
      jobStatus
      output
    }
  }
}
'''

get_notification = '''
  query getNotification($id: ID!){
    notification(id: $id){
      id
      title
      content
      context
      read
      metadata
      insertedAt
    }
  }
'''

get_group_users = '''
  query GetGroupUsers ($groupName: String!){
    groupUsers(filter: { groupName: $groupName }) {
      id
      group {
        id
        name
      }
      user {
        id
        name
        lastName
        email
        signInCount
        previousSignIp
        currentSignInIp
        previousSignInAt
        currentSignInAt
        insertedAt
      }
    }
  }
'''

get_worker_queues = '''
  query WorkerQueues($workerCode: String!){
    queueWorkers(filter: {workerCode: $workerCode}){
      id
      worker{
        id
        code
      }
      queue{
        id
        name
      }
      enabled
    }
  }
'''

get_queueLocks = '''
query(
  $workerCode: String
  $queueName: String
  $orderByQueueLock: OrderByQueueLock
) {
  queueLocks(
    orderBy: $orderByQueueLock
    filter: { workerCode: $workerCode, queueName: $queueName, active: true }
  ) {
    id
    active
  }
}
'''


import os
from pprint import pprint
from pygqlc import GraphQLClient
from .worker import ValiotWorker
from .worker import QueueType
# ! Create main helpers:
gql = GraphQLClient()
vw = ValiotWorker()
# * Initialize helpers:
gql.addEnvironment(
    'dev',
    url=os.environ.get('API'),
    wss=os.environ.get('WSS'),
    headers={'Authorization': os.environ.get('TOKEN')},
    default=True)
vw.setClient(gql)
vw.setWorker(os.environ.get('WORKER'))


@vw.job(
    name='TEST_JOB',
    alias='test job 1',
    description='',
    schedule='* * * * *',
    enabled=True,
    queueType=QueueType.FREQUENCY,
)
def test_job(job_id, update_job, kwargs):
    print("Hi, I'm test job #1")
    update_job({
        'id': job_id,
        'status': 'FINISHED',
        'output': '{}',
        'progress': 100
    })


@vw.job(
    name='TEST_JOB_2',
    alias='test job 2',
    description='',
    schedule='*/3 * * * *',
    enabled=True,
    queueType=QueueType.FREQUENCY,
)
def test_job_2(job_id, update_job, kwargs):
    print("Hi, I'm test job #2")
    update_job({
        'id': job_id,
        'status': 'FINISHED',
        'output': '{}',
        'progress': 100
    })


def main():
    print('main for valiot worker')
    vw.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function
import re
from time import time
import datetime
from dateutil.relativedelta import relativedelta
from dateutil.tz import tzutc
import calendar

step_search_re = re.compile(r'^([^-]+)-([^-/]+)(/(.*))?$')
search_re = re.compile(r'^([^-]+)-([^-/]+)(/(.*))?$')
only_int_re = re.compile(r'^\d+$')
any_int_re = re.compile(r'^\d+')
star_or_int_re = re.compile(r'^(\d+|\*)$')
VALID_LEN_EXPRESSION = [5, 6]


class CroniterError(ValueError):
    pass


class CroniterBadCronError(CroniterError):
    pass


class CroniterBadDateError(CroniterError):
    pass


class CroniterNotAlphaError(CroniterError):
    pass


class croniter(object):
    MONTHS_IN_YEAR = 12
    RANGES = (
        (0, 59),
        (0, 23),
        (1, 31),
        (1, 12),
        (0, 6),
        (0, 59)
    )
    DAYS = (
        31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
    )

    ALPHACONV = (
        {},
        {},
        {"l": "l"},
        {'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5, 'jun': 6,
         'jul': 7, 'aug': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12},
        {'sun': 0, 'mon': 1, 'tue': 2, 'wed': 3, 'thu': 4, 'fri': 5, 'sat': 6},
        {}
    )

    LOWMAP = (
        {},
        {},
        {0: 1},
        {0: 1},
        {7: 0},
        {},
    )

    bad_length = 'Exactly 5 or 6 columns has to be specified for iterator' \
                 'expression.'

    def __init__(self, expr_format, start_time=None, ret_type=float,
                 day_or=True):
        self._ret_type = ret_type
        self._day_or = day_or

        if start_time is None:
            start_time = time()

        self.tzinfo = None
        if isinstance(start_time, datetime.datetime):
            self.tzinfo = start_time.tzinfo
            # milliseconds/microseconds rounds
            if start_time.microsecond:
                start_time = start_time + relativedelta(seconds=1)
            start_time = self._datetime_to_timestamp(start_time)

        self.start_time = start_time
        self.dst_start_time = start_time
        self.cur = start_time

        self.expanded, self.nth_weekday_of_month = self.expand(expr_format)

    @classmethod
    def _alphaconv(cls, index, key, expressions):
        try:
            return cls.ALPHACONV[index][key.lower()]
        except KeyError:
            raise CroniterNotAlphaError(
                "[{0}] is not acceptable".format(" ".join(expressions)))

    def get_next(self, ret_type=None):
        return self._get_next(ret_type or self._ret_type, is_prev=False)

    def get_prev(self, ret_type=None):
        return self._get_next(ret_type or self._ret_type, is_prev=True)

    def get_current(self, ret_type=None):
        ret_type = ret_type or self._ret_type
        if issubclass(ret_type,  datetime.datetime):
            return self._timestamp_to_datetime(self.cur)
        return self.cur

    @classmethod
    def _datetime_to_timestamp(cls, d):
        """
        Converts a `datetime` object `d` into a UNIX timestamp.
        """
        if d.tzinfo is not None:
            d = d.replace(tzinfo=None) - d.utcoffset()

        return cls._timedelta_to_seconds(d - datetime.datetime(1970, 1, 1))

    def _timestamp_to_datetime(self, timestamp):
        """
        Converts a UNIX timestamp `timestamp` into a `datetime` object.
        """
        result = datetime.datetime.utcfromtimestamp(timestamp)
        if self.tzinfo:
            result = result.replace(tzinfo=tzutc()).astimezone(self.tzinfo)

        return result

    @classmethod
    def _timedelta_to_seconds(cls, td):
        """
        Converts a 'datetime.timedelta' object `td` into seconds contained in
        the duration.
        Note: We cannot use `timedelta.total_seconds()` because this is not
        supported by Python 2.6.
        """
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) \
            / 10**6

    # iterator protocol, to enable direct use of croniter
    # objects in a loop, like "for dt in croniter('5 0 * * *'): ..."
    # or for combining multiple croniters into single
    # dates feed using 'itertools' module
    def __iter__(self):
        return self
    __next__ = next = get_next

    def all_next(self, ret_type=None):
        '''Generator of all consecutive dates. Can be used instead of
        implicit call to __iter__, whenever non-default
        'ret_type' has to be specified.
        '''
        while True:
            yield self._get_next(ret_type or self._ret_type, is_prev=False)

    def all_prev(self, ret_type=None):
        '''Generator of all previous dates.'''
        while True:
            yield self._get_next(ret_type or self._ret_type, is_prev=True)

    iter = all_next  # alias, you can call .iter() instead of .all_next()

    def _get_next(self, ret_type=None, is_prev=False):
        expanded = self.expanded[:]
        nth_weekday_of_month = self.nth_weekday_of_month.copy()

        ret_type = ret_type or self._ret_type

        if not issubclass(ret_type, (float, datetime.datetime)):
            raise TypeError("Invalid ret_type, only 'float' or 'datetime' "
                            "is acceptable.")

        # exception to support day of month and day of week as defined in cron
        if (expanded[2][0] != '*' and expanded[4][0] != '*') and self._day_or:
            bak = expanded[4]
            expanded[4] = ['*']
            t1 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
            expanded[4] = bak
            expanded[2] = ['*']

            t2 = self._calc(self.cur, expanded, nth_weekday_of_month, is_prev)
            if not is_prev:
                result = t1 if t1 < t2 else t2
            else:
                result = t1 if t1 > t2 else t2
        else:
            result = self._calc(self.cur, expanded,
                                nth_weekday_of_month, is_prev)

        # DST Handling for cron job spanning accross days
        dtstarttime = self._timestamp_to_datetime(self.dst_start_time)
        dtstarttime_utcoffset = (
            dtstarttime.utcoffset() or datetime.timedelta(0))
        dtresult = self._timestamp_to_datetime(result)
        lag = lag_hours = 0
        # do we trigger DST on next crontab (handle backward changes)
        dtresult_utcoffset = dtstarttime_utcoffset
        if dtresult and self.tzinfo:
            dtresult_utcoffset = dtresult.utcoffset()
            lag_hours = (
                self._timedelta_to_seconds(dtresult - dtstarttime) / (60*60)
            )
            lag = self._timedelta_to_seconds(
                dtresult_utcoffset - dtstarttime_utcoffset
            )
        hours_before_midnight = 24 - dtstarttime.hour
        if dtresult_utcoffset != dtstarttime_utcoffset:
            if ((lag > 0 and lag_hours >= hours_before_midnight)
                    or (lag < 0 and
                        ((3600*lag_hours+abs(lag)) >= hours_before_midnight*3600))
                    ):
                dtresult = dtresult - datetime.timedelta(seconds=lag)
                result = self._datetime_to_timestamp(dtresult)
                self.dst_start_time = result
        self.cur = result
        if issubclass(ret_type, datetime.datetime):
            result = dtresult
        return result

    def _calc(self, now, expanded, nth_weekday_of_month, is_prev):
        if is_prev:
            nearest_diff_method = self._get_prev_nearest_diff
            sign = -1
            offset = (len(expanded) == 6 or now % 60 > 0) and 1 or 60
        else:
            nearest_diff_method = self._get_next_nearest_diff
            sign = 1
            offset = (len(expanded) == 6) and 1 or 60

        dst = now = self._timestamp_to_datetime(now + sign * offset)

        month, year = dst.month, dst.year
        current_year = now.year
        DAYS = self.DAYS

        def proc_month(d):
            if expanded[3][0] != '*':
                diff_month = nearest_diff_method(
                    d.month, expanded[3], self.MONTHS_IN_YEAR)
                days = DAYS[month - 1]
                if month == 2 and self.is_leap(year) is True:
                    days += 1

                reset_day = 1

                if diff_month is not None and diff_month != 0:
                    if is_prev:
                        d += relativedelta(months=diff_month)
                        reset_day = DAYS[d.month - 1]
                        d += relativedelta(
                            day=reset_day, hour=23, minute=59, second=59)
                    else:
                        d += relativedelta(months=diff_month, day=reset_day,
                                           hour=0, minute=0, second=0)
                    return True, d
            return False, d

        def proc_day_of_month(d):
            if expanded[2][0] != '*':
                days = DAYS[month - 1]
                if month == 2 and self.is_leap(year) is True:
                    days += 1
                if 'l' in expanded[2] and days == d.day:
                    return False, d

                if is_prev:
                    days_in_prev_month = DAYS[
                        (month - 2) % self.MONTHS_IN_YEAR]
                    diff_day = nearest_diff_method(
                        d.day, expanded[2], days_in_prev_month)
                else:
                    diff_day = nearest_diff_method(d.day, expanded[2], days)

                if diff_day is not None and diff_day != 0:
                    if is_prev:
                        d += relativedelta(
                            days=diff_day, hour=23, minute=59, second=59)
                    else:
                        d += relativedelta(
                            days=diff_day, hour=0, minute=0, second=0)
                    return True, d
            return False, d

        def proc_day_of_week(d):
            if expanded[4][0] != '*':
                diff_day_of_week = nearest_diff_method(
                    d.isoweekday() % 7, expanded[4], 7)
                if diff_day_of_week is not None and diff_day_of_week != 0:
                    if is_prev:
                        d += relativedelta(days=diff_day_of_week,
                                           hour=23, minute=59, second=59)
                    else:
                        d += relativedelta(days=diff_day_of_week,
                                           hour=0, minute=0, second=0)
                    return True, d
            return False, d

        def proc_day_of_week_nth(d):
            if '*' in nth_weekday_of_month:
                s = nth_weekday_of_month['*']
                for i in range(0, 7):
                    if i in nth_weekday_of_month:
                        nth_weekday_of_month[i].update(s)
                    else:
                        nth_weekday_of_month[i] = s
                del nth_weekday_of_month['*']

            candidates = []
            for wday, nth in nth_weekday_of_month.items():
                w = (wday + 6) % 7
                c = calendar.Calendar(w).monthdayscalendar(d.year, d.month)
                if c[0][0] == 0:
                    c.pop(0)
                for n in nth:
                    if len(c) < n:
                        continue
                    candidate = c[n - 1][0]
                    if (
                        (is_prev and candidate <= d.day) or
                        (not is_prev and d.day <= candidate)
                    ):
                        candidates.append(candidate)

            if not candidates:
                if is_prev:
                    d += relativedelta(days=-d.day,
                                       hour=23, minute=59, second=59)
                else:
                    days = DAYS[month - 1]
                    if month == 2 and self.is_leap(year) is True:
                        days += 1
                    d += relativedelta(days=(days - d.day + 1),
                                       hour=0, minute=0, second=0)
                return True, d

            candidates.sort()
            diff_day = (candidates[-1] if is_prev else candidates[0]) - d.day
            if diff_day != 0:
                if is_prev:
                    d += relativedelta(days=diff_day,
                                       hour=23, minute=59, second=59)
                else:
                    d += relativedelta(days=diff_day,
                                       hour=0, minute=0, second=0)
                return True, d
            return False, d

        def proc_hour(d):
            if expanded[1][0] != '*':
                diff_hour = nearest_diff_method(d.hour, expanded[1], 24)
                if diff_hour is not None and diff_hour != 0:
                    if is_prev:
                        d += relativedelta(
                            hours=diff_hour, minute=59, second=59)
                    else:
                        d += relativedelta(hours=diff_hour, minute=0, second=0)
                    return True, d
            return False, d

        def proc_minute(d):
            if expanded[0][0] != '*':
                diff_min = nearest_diff_method(d.minute, expanded[0], 60)
                if diff_min is not None and diff_min != 0:
                    if is_prev:
                        d += relativedelta(minutes=diff_min, second=59)
                    else:
                        d += relativedelta(minutes=diff_min, second=0)
                    return True, d
            return False, d

        def proc_second(d):
            if len(expanded) == 6:
                if expanded[5][0] != '*':
                    diff_sec = nearest_diff_method(d.second, expanded[5], 60)
                    if diff_sec is not None and diff_sec != 0:
                        d += relativedelta(seconds=diff_sec)
                        return True, d
            else:
                d += relativedelta(second=0)
            return False, d

        procs = [proc_month,
                 proc_day_of_month,
                 (proc_day_of_week_nth if nth_weekday_of_month
                     else proc_day_of_week),
                 proc_hour,
                 proc_minute,
                 proc_second]

        while abs(year - current_year) <= 1:
            next = False
            for proc in procs:
                (changed, dst) = proc(dst)
                if changed:
                    month, year = dst.month, dst.year
                    next = True
                    break
            if next:
                continue
            return self._datetime_to_timestamp(dst.replace(microsecond=0))

        if is_prev:
            raise CroniterBadDateError("failed to find prev date")
        raise CroniterBadDateError("failed to find next date")

    def _get_next_nearest(self, x, to_check):
        small = [item for item in to_check if item < x]
        large = [item for item in to_check if item >= x]
        large.extend(small)
        return large[0]

    def _get_prev_nearest(self, x, to_check):
        small = [item for item in to_check if item <= x]
        large = [item for item in to_check if item > x]
        small.reverse()
        large.reverse()
        small.extend(large)
        return small[0]

    def _get_next_nearest_diff(self, x, to_check, range_val):
        for i, d in enumerate(to_check):
            if d == "l":
                # if 'l' then it is the last day of month
                # => its value of range_val
                d = range_val
            if d >= x:
                return d - x
        return to_check[0] - x + range_val

    def _get_prev_nearest_diff(self, x, to_check, range_val):
        candidates = to_check[:]
        candidates.reverse()
        for d in candidates:
            if d != 'l' and d <= x:
                return d - x
        if 'l' in candidates:
            return -x
        candidate = candidates[0]
        for c in candidates:
            # fixed: c < range_val
            # this code will reject all 31 day of month, 12 month, 59 second,
            # 23 hour and so on.
            # if candidates has just a element, this will not harmful.
            # but candidates have multiple elements, then values equal to
            # range_val will rejected.
            if c <= range_val:
                candidate = c
                break

        return (candidate - x - range_val)

    def is_leap(self, year):
        if year % 400 == 0 or (year % 4 == 0 and year % 100 != 0):
            return True
        else:
            return False

    @classmethod
    def expand(cls, expr_format):
        expressions = expr_format.split()

        if len(expressions) not in VALID_LEN_EXPRESSION:
            raise CroniterBadCronError(cls.bad_length)

        expanded = []
        nth_weekday_of_month = {}

        for i, expr in enumerate(expressions):
            e_list = expr.split(',')
            res = []

            while len(e_list) > 0:
                e = e_list.pop()

                if i == 4:
                    e, sep, nth = str(e).partition('#')
                    if nth and not re.match(r'[1-5]', nth):
                        raise CroniterBadDateError(
                            "[{0}] is not acceptable".format(expr_format))

                t = re.sub(r'^\*(\/.+)$', r'%d-%d\1' % (
                    cls.RANGES[i][0],
                    cls.RANGES[i][1]),
                    str(e))
                m = search_re.search(t)

                if not m:
                    t = re.sub(r'^(.+)\/(.+)$', r'\1-%d/\2' % (
                        cls.RANGES[i][1]),
                        str(e))
                    m = step_search_re.search(t)

                if m:
                    (low, high, step) = m.group(1), m.group(2), m.group(4) or 1

                    if not any_int_re.search(low):
                        low = "{0}".format(cls._alphaconv(i, low, expressions))

                    if not any_int_re.search(high):
                        high = "{0}".format(
                            cls._alphaconv(i, high, expressions))

                    if (
                        not low or not high or int(low) > int(high)
                        or not only_int_re.search(str(step))
                    ):
                        if i == 4 and high == '0':
                            # handle -Sun notation -> 7
                            high = '7'
                        else:
                            raise CroniterBadDateError(
                                "[{0}] is not acceptable".format(expr_format))

                    low, high, step = map(int, [low, high, step])
                    try:
                        rng = range(low, high + 1, step)
                    except ValueError as exc:
                        raise CroniterBadCronError(
                            'invalid range: {0}'.format(exc))
                    e_list += (["{0}#{1}".format(item, nth) for item in rng]
                               if i == 4 and nth else rng)
                else:
                    if t.startswith('-'):
                        raise CroniterBadCronError(
                            "[{0}] is not acceptable,\
                            negative numbers not allowed".format(
                                expr_format))
                    if not star_or_int_re.search(t):
                        t = cls._alphaconv(i, t, expressions)

                    try:
                        t = int(t)
                    except:
                        pass

                    if t in cls.LOWMAP[i]:
                        t = cls.LOWMAP[i][t]

                    if (
                        t not in ["*", "l"]
                        and (int(t) < cls.RANGES[i][0] or
                             int(t) > cls.RANGES[i][1])
                    ):
                        raise CroniterBadCronError(
                            "[{0}] is not acceptable, out of range".format(
                                expr_format))

                    res.append(t)

                    if i == 4 and nth:
                        if t not in nth_weekday_of_month:
                            nth_weekday_of_month[t] = set()
                        nth_weekday_of_month[t].add(int(nth))

            res.sort()
            expanded.append(['*'] if (len(res) == 1
                                      and res[0] == '*')
                            else res)

        return expanded, nth_weekday_of_month

    @classmethod
    def is_valid(cls, expression):
        try:
            cls.expand(expression)
        except CroniterError:
            return False
        else:
            return True



from datetime import datetime
from pytz import timezone
from tzlocal import get_localzone


def getUtcDate(date: datetime) -> datetime:
    utc = timezone('UTC')
    tz = get_localzone()  # local timezone
    local_date = date.replace(tzinfo=tz)
    utc_date = local_date.astimezone(utc)
    return utc_date



from .worker import ValiotWorker
from .worker import QueueType, JobStatus, PollingMode, JobConfigMode, QueryOrderBy
from .uploaders import update_job

# * Package name:
name = 'ValiotWorker'
# * required here for pypi upload exceptions:
__version__ = "4.2.0"



create_job = '''
  mutation createJob(
    $queueName: String!
    $worker: String
  ){
    createJob(
      queueName: $queueName
      workerCode: $worker
      jobStatus: WAITING
      progress: 0
      input: "{}"
      output: "{}"
    ){
      successful
      messages{
        message
        field
      }
      result{
        id
        input
        output
        jobStatus
        progress
        queue{
          id
          name
          updatedAt
        }
      }
    }
  }
'''

update_job = '''
  mutation UpdateJob(
    $id: ID!
    $status: JobStatus!
    $output: Jsonb
    $progress: Int
  ) {
    updateJob(
      id: $id
      job: {
        jobStatus: $status
        progress: $progress
        output: $output
      }
    ) {
      successful
      result {
        id
      }
      messages {
        message
      }
    }
  }
'''

delete_job = '''
  mutation abortJob($id: ID!) {
    updateJob(id: $id job: { jobStatus: ABORTED }) {
      successful
      messages{
        field
        message
      }
      result{
        id
      }
    }
  }
'''

upsert_queue = '''
mutation upsertQueue(
  $name: String!
  $alias: String!
  $type: QueueType!
  $enabled: Boolean
  $description: Text
  $schedule: String
  $query: Text
  $lockRequired: Boolean
) {
  upsertQueue(
    name: $name
    queue: {
      alias: $alias
      type: $type
      enabled: $enabled
      description: $description
      schedule: $schedule
      query: $query
      lockRequired: $lockRequired
    }
  ) {
    successful
    result {
      id
      name
      alias
      type
      description
      schedule
      enabled
      query
      lockRequired
    }
    messages {
      message
      field
    }
  }
}
'''

run_job = '''
  mutation runJob(
    $id: ID!
    $status: JobStatus!
    $queueName: String!, 
    $lastRunAt: DateTime!
  ) {
    updateJob(
      id: $id
      job: {
        jobStatus: $status
        progress: 0
        output: "{}"
        }
    ) {
      successful
      messages { field message }
      result { id jobStatus progress }
    }
    updateQueue(
      findBy: { name: $queueName },
      queue: { lastRunAt: $lastRunAt }
    ) {
      successful
      messages { field message }
      result { id name lastRunAt }
    }
  }
'''

run_event_job = '''
  mutation runJob($queueName: String!, $lastRunAt: DateTime!) {
  job: createJob(
    queueName: $queueName
    jobStatus: RUNNING
    progress: 0
    output: "{}"
  ) {
    successful
    messages {
      field
      message
    }
    result {
      id
      jobStatus
      progress
    }
  }
  queue: updateQueue(
    findBy: { name: $queueName }
    queue: { lastRunAt: $lastRunAt }
  ) {
    successful
    messages {
      field
      message
    }
    result {
      id
      name
      lastRunAt
    }
  }
}
'''

create_notification = '''
mutation CreateNotification(
  $userId: ID
  $context: NotificationType,
  $title: String,
  $content: Text!,
  $link: String,
  $metadata: String,
  $linkText: String
){
  createNotification(
    read: false
    userId: $userId
    context: $context
    title: $title
    content: $content
    link: $link
    linkText: $linkText
    metadata: $metadata
  ){
    successful
    result{
      id
      title
      content
      context
      metadata
      insertedAt
      updatedAt
    }
    messages{
      message
    }
  }
}
'''

'''
Queue metadata:
{
  "reminder": {
    "id": "123"
    "SentAt": "2019-07-26T12:51:16Z",
    # ! The reminder is sent without a RESOLVED prop
    "enabled": true,
  },
  "notification": {
    "id": "121"
    "title": "blabla",
    "content": "this is notification",
    "context": "DANGER",
    "metadata": "..." # ! If this function is used to create notifications, we assume a "resolved: false" must be sent with it
  }
}
'''
update_queue_meta = '''
  mutation updateQueueMetadata(
    $queueName: String!
    $metadata: Text!
  ){
    updateQueue(
      findBy:{name: $queueName}
      queue:{
        query: $metadata
      }
    ){
      successful
      messages{
        message
        field
      }
      result{
        id
        name
        alias
        description
        enabled
        schedule
        query
      }
    }
  }
'''

update_notification_metadata = '''
  mutation updateMetadata(
    $id: ID!
    $metadata: String!
  ){
    updateNotification(
      id:$id
      notification:{
        metadata: $metadata
      }
    ){
      successful
      messages{
        field
        message
      }
      result{
        id
        read
        title
        content
        context
        metadata
        insertedAt
        updatedAt
      }
    }
  }
'''

create_lock = '''
mutation createLock(
  $worker: String!
  $queue: String!
  $startAt: DateTime
){
  createQueueLock(
    active: true
    workerCode: $worker
    queueName: $queue
    startAt: $startAt
  ){
    successful
    messages{
      field
      message
    }
    result{
      id
      active
      startAt
      endAt
      insertedAt
      worker{id code}
      queue{id name}
    }
  }
}
'''

update_lock = '''
mutation UpdateLock($id: ID!, $active: Boolean!, $endAt: DateTime) {
  updateQueueLock(id: $id, queueLock: { active: $active, endAt: $endAt }) {
    successful
    messages {
      field
      message
    }
    result {
      id
      active
      startAt
      endAt
      insertedAt
      worker {
        id
        code
      }
      queue {
        id
        name
      }
    }
  }
}
'''

update_context = '''
mutation UpdateContext(
  $queue: String!
  $context: Jsonb!
){
  updateQueue(
    findBy:{ name: $queue }
    queue: {
      context: $context
    }
  ){
    successful
    messages{
      field
      message
    }
    result{
      id
      name
      context
    }
  }
}
'''

create_queue_worker = '''
  mutation (
    $worker: String!
    $queue: String!
    $enabled: Boolean!
  ){
    createQueueWorker(
      workerCode: $worker
      queueName: $queue
      enabled: $enabled
    ) {
      successful
      messages {
        field
        message
      }
    }
  }
'''

update_queue_worker = '''
  mutation (
    $id: ID!
    $enabled: Boolean!
  ){
    updateQueueWorker(
      id: $id
      queueWorker: {
        enabled: $enabled
      }
    ) {
      successful
      messages {
        field
        message
      }
    }
  }
'''


def build_batch_stale_mutation(jobs):
    batch_template = '''mutation DeleteStaleJobs{{\n\t{mutation_items}\n}}'''
    item_template = '''{label}: updateJob(id: {id} job: {{ jobStatus: ABORTED }}) {{ successful messages{{field message}} }}'''
    mutation_items = [
        item_template.format(label=f'delete_job_{job["id"]}', id=job["id"])
        for job in jobs
    ]
    return (
        batch_template.format(mutation_items='\n\t'.join(mutation_items))
        if jobs
        else None
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from ValiotWorker import __version__

requirements = [
    # TODO: put package requirements here
    'pygqlc>=3.0.1',
    'pydash',
    'pytz',
    'tzlocal',
    'colorama',
    'termcolor',
    'python-dateutil',
    'singleton-decorator',
    'redis',
]

setup_requirements = [
    # TODO(alanbato): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    # TODO: put package test requirements here
    'pytest',
    'pytest-cov',
]

desc = "Enables running Python functions as standalone jobs based on interaction with valiot-jobs API"
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ValiotWorker',
    version=__version__,
    description=desc,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Valiot",
    author_email="hiring@valiot.io",
    url='https://github.com/valiot/python-gql-worker',
    packages=find_packages(include=['ValiotWorker']),
    entry_points={
        'console_scripts': [
            'ValiotWorker=ValiotWorker.__main__:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='ValiotWorker',
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



from .croniter import croniter
from .Logging import log, LogLevel


def get_croniter(schedule, date):
    cron_iter = None
    try:
        cron_iter = croniter(schedule, date)
    except Exception as e:
        log(LogLevel.ERROR, f'croniter error: {e}')
    return cron_iter
